import xml.etree.ElementTree as ET
import xml.dom.minidom as minidom
from datetime import datetime

from gamuLogger import Logger

Logger.setModule("melkor")


from ..customTypes import Suite, Test, TestList


def formatDate(date: datetime) -> str:
    #return the date in the ISO 8601 format
    return date.strftime("%Y-%m-%dT%H:%M:%S")

class Report:
    def __init__(self, TestList: TestList):
        self.__testList = TestList
    
    def __generateTestElement(self, test: Test, className : str) -> ET.Element:
        Logger.debug(f"Generating test element for {test.name}")
        element = ET.Element("testcase")
        element.set("name", test.name)
        element.set("classname", className)
        element.set("time", str(test.time))
        if test.failed:
            failure = ET.Element("failure")
            failure.set("message", test.message)
            failure.set("type", test.exceptionType)
            failure.text = test.traceback
            element.append(failure)
        if test.error:
            error = ET.Element("error")
            error.set("message", test.message)
            error.set("type", test.exceptionType)
            error.text = test.traceback
            element.append(error)
        if test.skipped:
            skipped = ET.Element("skipped")
            skipped.set("message", test.message)
            element.append(skipped)
        
        output = test.output
        if output:

            systemOut = ET.Element("system-out")
            systemOut.text = '\n'.join(output)
            element.append(systemOut)
            
        if test.doc:
            doc = ET.Element("property")
            doc.set("name", "description")
            doc.set("value", test.doc)
            element.append(doc)
        
        Logger.debug(f"Generated test element for {test.name}")
        return element
    
    def __generateSuiteElement(self, suite: Suite) -> ET.Element:
        Logger.debug(f"Generating suite element for {suite.name}")
        element = ET.Element("testsuite")
        element.set("name", str(suite.name))
        element.set("tests", str(suite.tests))
        element.set("failures", str(suite.failures))
        element.set("errors", str(suite.errors))
        element.set("skipped", str(suite.skipped))
        element.set("time", str(suite.time))
        element.set("timestamp", formatDate(suite.timestamp))
        element.set("file", suite.file)
        for child in suite.childs.values():
            if isinstance(child, Suite):
                element.append(self.__generateSuiteElement(child))
            else:
                element.append(self.__generateTestElement(child, suite.name))
        return element
    
    def __generateSuitesElement(self, testList: TestList) -> ET.Element:
        Logger.debug(f"Generating suites element for {testList.name}")
        element = ET.Element("testsuites")
        element.set("name", str(testList.name))
        element.set("tests", str(testList.tests))
        element.set("failures", str(testList.failures))
        element.set("errors", str(testList.errors))
        element.set("skipped", str(testList.skipped))
        element.set("time", str(testList.time))
        element.set("timestamp", formatDate(testList.timestamp))
        for child in testList.childs.values():
            if isinstance(child, Suite):
                element.append(self.__generateSuiteElement(child))
            else:
                element.append(self.__generateTestElement(child, testList.name))
        return element
    
    def generate(self) -> str:
        root = self.__generateSuitesElement(self.__testList)
        return minidom.parseString(ET.tostring(root)).toprettyxml()
    
    def save(self, path: str):
        with open(path, 'w') as file:
            file.write(self.generate())
            



import datetime
from typing import Callable, Dict, Any, Union

from .utils import Chrono


class Test:
    def __init__(self, func : Callable[[], Dict[str, Any]], parent : Union['TestList', 'Suite'] = None):
        self.__func = func
        self.__name = func.__annotations__["name"]
        self.__failed = False
        self.__error = False
        self.__skipped = False
        self.__time = 0.0
        self.__message = ""
        self.__traceback = None
        self.__output = None
        self.__parent = parent
        
    def run(self):
        if self.__skipped:
            return # Do not run the test if it was marked as skipped
        chrono = Chrono()
        with chrono:
            data = self.__func() # contains keys: error, failed, exception, traceback, output
        self.__time = chrono.get()
        self.__failed = data['failed']
        self.__error = data['error']
        self.__exceptionType = data["exception"].__class__.__name__ if data["exception"] is not None else ""
        self.__message = data["exception"].__str__() if data["exception"] is not None else ""
        self.__traceback = data["traceback"]
        self.__output = data["output"]
        
        if self.__parent is not None:
            self.__parent.update()
            
    def completeName(self):
        if self.__parent is None:
            return self.__name
        return self.__parent.completeName() + "." + self.__name

        
    def skip(self, message : str):
        self.__skipped = True
        self.__message = message
        
    @property
    def name(self):
        return self.__name
    
    @property
    def failed(self):
        return self.__failed
    
    @property
    def error(self):
        return self.__error
    
    @property
    def skipped(self):
        return self.__skipped
    
    @property
    def time(self):
        return self.__time  
    
    @property
    def message(self):
        return self.__message
    
    @property
    def traceback(self):
        return self.__traceback
    
    @property
    def output(self):
        return self.__output
    
    @property
    def exceptionType(self):
        return self.__exceptionType
    
    @property
    def doc(self):
        return self.__func.__doc__
    
    def __str__(self):
        return f"{self.name} - {self.time} seconds - {'Failed' if self.failed else 'Succeeded'} - {'Error' if self.error else 'No error'} - {'Skipped' if self.skipped else 'Not skipped'}"
    
    

TESTLIST = None

class TestList:
    def __init__(self, name : str, parent : Union['TestList', 'Suite'] = None):
        self._timestamp = datetime.datetime.now() # Date and time of when the test run was executed
        self._name = name
        self._tests = 0
        self._failures = 0
        self._errors = 0
        self._skipped = 0
        self._time = 0.0
        self._childs = {} #type: Dict[str, Union[Test, Suite]]
        self._parent = parent
        
    def hasChild(self, name : str):
        return name in self._childs
    
    def getChild(self, name : str):
        return self._childs[name]
        
    def addTest(self, test : Test):
        self._childs[test.name] = test
        self._tests += 1
        self._failures += 1 if test.failed else 0
        self._errors += 1 if test.error else 0
        self._skipped += 1 if test.skipped else 0
        self._time += test.time
        
    def addSuite(self, suite: 'Suite'):
        self._childs[suite.name] = suite
        self._tests += suite.tests
        self._failures += suite.failures
        self._errors += suite.errors
        self._skipped += suite.skipped
        self._time += suite.time
        
    def run(self):
        for child in self._childs.values():
            child.run()
    
    def update(self):
        self._tests = 0
        self._failures = 0
        self._errors = 0
        self._skipped = 0
        self._time = 0.0
        for child in self._childs.values():
            if isinstance(child, Test):
                self._tests += 1
                self._failures += 1 if child.failed else 0
                self._errors += 1 if child.error else 0
                self._skipped += 1 if child.skipped else 0
                self._time += child.time
            else:
                self._tests += child.tests
                self._failures += child.failures
                self._errors += child.errors
                self._skipped += child.skipped
                self._time += child.time
        
        if self._parent is not None:
            self._parent.update()
            
    def completeName(self):
        if self._parent is None:
            return self._name
        return self._parent.completeName() + "." + self._name
        
    @property
    def timestamp(self):
        return self._timestamp
    
    @property
    def name(self):
        return self._name
    
    @property
    def tests(self):
        return self._tests
    
    @property
    def failures(self):
        return self._failures
    
    @property
    def errors(self):
        return self._errors
    
    @property
    def skipped(self):
        return self._skipped
    
    @property
    def time(self):
        return self._time
    
    @property
    def childs(self):
        return self._childs        

    @staticmethod
    def getInstance():
        global TESTLIST
        if TESTLIST is None:
            raise Exception("TestList instance not created")
        return TESTLIST
    
    @staticmethod
    def new(name : str):
        global TESTLIST
        TESTLIST = TestList(name)
        return TESTLIST
    
    def __str__(self):
        return "TestList : " + self.name + "\n" + "\n".join([str(child) for child in self.childs.values()])
    
    def getHierarchy(self) -> Dict[str, Union[Dict, str]]: # Returns a dictionary with the hierarchy of it's children
        result = {}
        for child in self.childs.values():
            if isinstance(child, Test):
                result[child.name] = child.name
            else:
                result[child.name] = child.getHierarchy()
        return result


class Suite(TestList):
    def __init__(self, file : str, name : str = None, parent : Union['TestList', 'Suite'] = None):
        super().__init__(name if name is not None else file, parent)
        self._file = file
        
    @property
    def file(self):
        return self._file
    
    def __str__(self):
        return f"Suite {self.name} ({self.file}) : {self.tests} tests, {self.failures} failures, {self.errors} errors, {self.skipped} skipped, {self.time} seconds"


class ReturnCodeError(Exception):
    pass


from gamuLogger import Logger
import inspect
import os

from .utils import CapturePrint, formatTraceback
from .customTypes import TestList, Test, Suite, ReturnCodeError
from .settings import Settings


Logger.setModule("melkor")

Settings.setFilePath("config.json")

 
class _testclassdecoratorBase:
	def __init__(self, func):
		self.__func = func
		self.__call__.__annotations__["name"] = func.__name__
  
		filepath = os.path.relpath(inspect.getfile(func), Settings().get("testDir"))
		suiteHierarchy = filepath.replace(".py", "").split(os.sep)
		crtSuite = TestList.getInstance()
		for suiteName in suiteHierarchy:
			if not crtSuite.hasChild(suiteName):
				Logger.debug(f"Creating suite {suiteName}")
				crtSuite.addSuite(Suite(filepath, suiteName, crtSuite))
			crtSuite = crtSuite.getChild(suiteName)
   
		crtSuite.addTest(Test(self.__call__, crtSuite))
		Logger.debug(f"Registered test {self.__func.__name__}")
  
	def __call__(self, *args, **kwargs):
		Logger.debug(f"Running test {self.__func.__name__}")
   
		exitData = {
			"error": False,
			"failed": False,
			"exception": None,
			"traceback": None,
		}
		result = None
		try:
			capture = CapturePrint()
			with capture:
				result = self.__func(*args, **kwargs)
		except AssertionError as e:
			exitData["failed"] = True
			exitData["exception"] = e
			exitData["traceback"] = formatTraceback(e)
		except Exception as e:
			exitData["error"] = True
			exitData["exception"] = e
			exitData["traceback"] = formatTraceback(e)

		if type(result) == int and result != 0:
			exitData["failed"] = True
			exitData["exception"] = ReturnCodeError(f"Test returned {result}")
   
		exitData["output"] = capture.get()
		return exitData


def UnitTest(allowedToFail=False):
	class _unittestclassdecorator(_testclassdecoratorBase):
		def __init__(self, func):
			super().__init__(func)
			
		def __call__(self, *args, **kwargs):
			data = super().__call__(*args, **kwargs)
			data["allowedToFail"] = allowedToFail

			return data

	return _unittestclassdecorator



import importlib.util
from typing import List, Callable
import os

from gamuLogger import Logger

from .customTypes import TestList
from .settings import Settings


Logger.setModule("melkor")

def importFile(file) -> object:
    spec = importlib.util.spec_from_file_location("module.name", file)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module

def importFiles(files : List[str]) -> List[object]:
    result = []
    for file in files:
        if not os.path.exists(file):
            raise FileNotFoundError(f"File '{file}' not found")
        Logger.debug(f"Importing file '{file}'")
        result.append(importFile(file))
    return result


def runTests(testList : List[Callable[[], int]]) -> TestList:
    result = TestList(Settings.get("name"))
    
    for test in testList:
        data = test()
        if data["hasSucceeded"]:
            result['failedTests'] += 1
            string = f"Test {test.__annotations__['name']} failed"
            if data["exception"]:
                string += f" with exception: {data['exception']}"
            if data["output"]:
                string += "\nOutput:\n    " + '\n    '.join(data['output'])
            if data["traceback"]:
                string += "\nStack:\n    " + data['traceback']
            Logger.error(string)
            
            if not data["allowedToFail"]:
                result['failed'] = True
        
        else:
            Logger.debug(f"Test {test.__annotations__['name']} succeeded")
    
    return result



import json5 as json
import os

class Settings:
    """
    Read-only settings class that reads a JSON file and provides access
    """
    __instance = None
    __configFile = 'melkor.json'
    
    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
        return cls.__instance
    
    def __init__(self):
        self.__data = {}
        self.load()
        
    def load(self):
        if not os.path.exists(Settings.__configFile):
            raise FileNotFoundError(f"Settings file '{Settings.__configFile}' not found")
        with open(Settings.__configFile, 'r') as file:
            self.__data = json.load(file)
            
    def get(self, key : str, default = None):
        tokens = key.split('.')
        data = self.__data
        try:
            for token in tokens:
                data = data[token]
            return data
        except KeyError:
            if default is not None:
                return default
            raise KeyError(f"Key '{key}' not found in settings")
            
    
    def __getitem__(self, key : str):
        return self.get(key)
    
    def __contains__(self, key : str):
        try:
            self.get(key)
            return True
        except KeyError:
            return False
        
    def has(self, key : str):
        return key in self
    
    @staticmethod
    def setFilePath(path : str):
        Settings.__configFile = path
        
        
            
        
    


import argparse
import sys
import os
import importlib.metadata as metadata
import traceback


from gamuLogger import Logger, LEVELS

from .settings import Settings
from .engine import importFiles
from .customTypes import TestList
from .output.junit import Report as JunitReport

Logger.setModule("melkor")

def logPackageVersion():
    Logger.debug(f"Python version: {sys.version}")
    Logger.debug(f"gamuLogger version: {metadata.version('gamuLogger')}")
    Logger.debug(f"melkor version: {metadata.version('melkor')}")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("configFile", help="Path to the configuration file")
    parser.add_argument("--debug", help="Enable debug mode", action="store_true")
    args = parser.parse_args()
    
    if args.debug:
        Logger.setLevel('stdout', LEVELS.DEBUG)
        
    logPackageVersion()
    
    Settings.setFilePath(args.configFile)
        
    testDir = Settings().get("testDir")
    if not os.path.exists(testDir):
        Logger.error(f"Test directory '{testDir}' not found")
        sys.exit(1)

    TestList.new(Settings().get("name"))

    files = [os.path.join(testDir, file) for file in os.listdir(testDir) if file.endswith(".py")]
    Logger.info(f"Found {len(files)} test files, loading them")
    modules = importFiles(files)
    
    Logger.info("Running tests")
    TestList.getInstance().run()
    
    Logger.info("Generating JUnit report")
    junitReport = JunitReport(TestList.getInstance())
    junitReport.save(Settings().get("outFile"))
    Logger.info(f"Report generated to {Settings().get('outFile')}")


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        Logger.debug(traceback.format_exc())
        Logger.critical(f"An exception occurred: {e}")



import builtins
import traceback
import time


class CapturePrint:
    def __init__(self):
        self.__stdout = [] #type: list[str]
        
    def __enter__(self):
        self.__buildinPrint = builtins.print
        builtins.print = self.__print
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        builtins.print = self.__buildinPrint
        
    def __print(self, *args, **kwargs):
        self.__stdout.append(" ".join(args))
        
    def get(self) -> list[str]:
        return self.__stdout
    
    
def formatTraceback(exception: Exception) -> str:
    """Format the traceback of an exception, removing the first element"""
    trace =  traceback.TracebackException.from_exception(exception)
    trace.stack.pop(0)
    return "    ".join(trace.format())


class Chrono:
    def __init__(self):
        self.__start = None
        self.__end = None
    
    def __enter__(self):
        self.__start = time.time()
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__end = time.time()
        
    def get(self):
        if self.__end is None: # Get the value of the chrono, without stopping it
            return time.time() - self.__start
        return self.__end - self.__start
